
# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: era_5g_client:dist

from setuptools import setup

setup(**{
    'author': 'Michal Kapinus',
    'author_email': 'ikapinus@fit.vutbr.cz',
    'classifiers': [
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
    ],
    'description': 'A client for 5G-ERA NetApps',
    'install_requires': (
        'era-5g-interface~=0.4.1',
        'numpy>=1.24.2',
        'opencv-python>=4.6.0.66',
        'python-socketio[client]>=5.8.0',
        'requests>=2.28.2',
        'types-requests==2.31.0.1',
    ),
    'license': 'LGPL',
    'name': 'era_5g_client',
    'namespace_packages': (
    ),
    'package_data': {
        'era_5g_client': (
            'py.typed',
        ),
    },
    'packages': (
        'era_5g_client',
    ),
    'python_requires': '>=3.8',
    'version': '0.4.2',
})
